/*
 *  COPYRIGHT (c) 2009 Chris Johns <chrisj@rtems.org>
 *
 *  COPYRIGHT (c) 1989-2009. On-Line Applications Research Corporation (OAR).
 *
 *  The license and distribution terms for this file may be
 *  found in the file LICENSE in this distribution or at
 *  http://www.rtems.com/license/LICENSE.
 *
 *  $Id$
 */
/**
 * @file
 *
 * @ingroup rtems-ext2fs
 *
 * RTEMS Block IO Manager For EXT2FS
 */

#define RTEMS_EXT2FS_BLOCKIO_DEBUG 0

#if HAVE_CONFIG_H
#include "config.h"
#endif

#include <unistd.h>
#include <stdlib.h>

#include "rtems-ext2fs-private.h"

#include <rtems/diskdevs.h>
#include <rtems/bdbuf.h>

/*
 * For checking structure magic numbers...
 */

#define EXT2_ET_MAGIC_RTEMS_BLOCKIO_CHANNEL (0x28092009)
#define EXT2_CHECK_MAGIC(struct, code) \
  if ((struct)->magic != (code)) return (code)

struct rtems_ext2fs_blockio_private_data {
  int                    magic;
  dev_t                  dev;
  rtems_disk_device*     dd;
  int                    access_time;
  struct struct_io_stats io_stats;
};

static errcode_t rtems_ext2fs_blockio_open(const char *name, int flags, io_channel *channel);
static errcode_t rtems_ext2fs_blockio_close(io_channel channel);
static errcode_t rtems_ext2fs_blockio_set_blksize(io_channel channel, int blksize);
static errcode_t rtems_ext2fs_blockio_read_blk(io_channel channel, unsigned long block,
                                               int count, void *data);
static errcode_t rtems_ext2fs_blockio_write_blk(io_channel channel, unsigned long block,
                                                int count, const void *data);
static errcode_t rtems_ext2fs_blockio_flush(io_channel channel);
static errcode_t rtems_ext2fs_blockio_set_option(io_channel channel, const char *option,
                                                 const char *arg);
static errcode_t rtems_ext2fs_blockio_get_stats(io_channel channel, io_stats *stats);
static errcode_t rtems_ext2fs_blockio_read_blk64(io_channel channel, unsigned long long block,
                                                 int count, void *data);
static errcode_t rtems_ext2fs_blockio_write_blk64(io_channel channel, unsigned long long block,
                                                  int count, const void *data);

static struct struct_io_manager struct_rtems_ext2fs_blockio_manager = {
  EXT2_ET_MAGIC_IO_MANAGER,
  "RTEMS Block I/O Manager",
  rtems_ext2fs_blockio_open,
  rtems_ext2fs_blockio_close,
  rtems_ext2fs_blockio_set_blksize,
  rtems_ext2fs_blockio_read_blk,
  rtems_ext2fs_blockio_write_blk,
  rtems_ext2fs_blockio_flush,
  0,
  rtems_ext2fs_blockio_set_option,
  rtems_ext2fs_blockio_get_stats,
  rtems_ext2fs_blockio_read_blk64,
  rtems_ext2fs_blockio_write_blk64,
};

io_manager rtems_ext2fs_blockio_manager = &struct_rtems_ext2fs_blockio_manager;

static errcode_t rtems_ext2fs_blockio_get_stats(io_channel channel,
                                                io_stats  *stats)
{
  errcode_t retval = 0;

  struct rtems_ext2fs_blockio_private_data *data;

  EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
  data = (struct rtems_ext2fs_blockio_private_data *) channel->private_data;
  EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_RTEMS_BLOCKIO_CHANNEL);

  if (stats)
    *stats = &data->io_stats;

  return retval;
}

static errcode_t rtems_ext2fs_blockio_open(const char *name,
                                           int         flags,
                                           io_channel *channel)
{
  io_channel                                io = NULL;
  struct rtems_ext2fs_blockio_private_data *data = NULL;
  errcode_t                                 retval;
  struct stat                               st;

  if (name == 0)
    return EXT2_ET_BAD_DEVICE_NAME;
  retval = ext2fs_get_mem(sizeof(struct struct_io_channel), &io);
  if (retval)
    return retval;
  memset(io, 0, sizeof(struct struct_io_channel));
  io->magic = EXT2_ET_MAGIC_IO_CHANNEL;
  retval = ext2fs_get_mem(sizeof(struct rtems_ext2fs_blockio_private_data), &data);
  if (retval)
    goto cleanup;

  io->manager = rtems_ext2fs_blockio_manager;
  retval = ext2fs_get_mem(strlen(name)+1, &io->name);
  if (retval)
    goto cleanup;

  strcpy(io->name, name);
  io->private_data = data;
  io->block_size = 512;
  io->read_error = 0;
  io->write_error = 0;
  io->refcount = 1;

  memset(data, 0, sizeof(struct rtems_ext2fs_blockio_private_data));
  data->magic = EXT2_ET_MAGIC_RTEMS_BLOCKIO_CHANNEL;
  data->io_stats.num_fields = 2;

  if (stat(name, &st) < 0)
  {
    retval = errno;
    goto cleanup;
  }

  if (!S_ISCHR(st.st_mode))
  {
    retval = EIO;
    goto cleanup;
  }

  /* check that device is registred as block device and lock it */
  data->dev = st.st_rdev;
  data->dd = rtems_disk_obtain(data->dev);
  if (data->dd == NULL)
  {
    retval = EIO;
    goto cleanup;
  }

#if RTEMS_EXT2FS_BLOCKIO_DEBUG
  printf("rtems-ext2fs: blockio: opening:%s ddev:%d\n", name, data->dev);
#endif
  
  *channel = io;
  return 0;

cleanup:
  if (data) {
    ext2fs_free_mem(&data);
  }
  if (io)
    ext2fs_free_mem(&io);
  return retval;
}

static errcode_t rtems_ext2fs_blockio_close(io_channel channel)
{
  struct rtems_ext2fs_blockio_private_data *data;
  errcode_t                                 retval = 0;

  EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
  data = (struct rtems_ext2fs_blockio_private_data *) channel->private_data;
  EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_RTEMS_BLOCKIO_CHANNEL);

  if (--channel->refcount > 0)
    return 0;

  rtems_bdbuf_syncdev(data->dev);
  rtems_disk_release(data->dd);

  ext2fs_free_mem(&channel->private_data);
  if (channel->name)
    ext2fs_free_mem(&channel->name);
  ext2fs_free_mem(&channel);
  return retval;
}

static errcode_t rtems_ext2fs_blockio_set_blksize(io_channel channel,
                                                  int        blksize)
{
  struct rtems_ext2fs_blockio_private_data *data;
  errcode_t    retval = 0;

  EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
  data = (struct rtems_ext2fs_blockio_private_data *) channel->private_data;
  EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_RTEMS_BLOCKIO_CHANNEL);

  /*
   * We have the libblock cache under us so this does not
   * mean too much.
   */
  channel->block_size = blksize;

  return retval;
}

static errcode_t rtems_ext2fs_blockio_read_blk64(io_channel         channel,
                                                 unsigned long long block,
                                                 int                count,
                                                 void               *buf)
{
  struct rtems_ext2fs_blockio_private_data *data;
  char                                     *cp;
  ssize_t                                   size;
  int                                       actual = 0;
  errcode_t                                 retval = 0;
  rtems_blkdev_bnum                         bdblock;

  EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
  data = (struct rtems_ext2fs_blockio_private_data *) channel->private_data;
  EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_RTEMS_BLOCKIO_CHANNEL);

  /*
   * Handle the strange count value. If -ve the count is in bytes
   * and if +ve the count is in the block size.
   */
  size = (count < 0) ? -count : count * channel->block_size;

  /*
   * Map the block number from the channel block to the bdbuf block
   * number.
   */
  bdblock = (block * channel->block_size) / data->dd->block_size;
  
#if RTEMS_EXT2FS_BLOCKIO_DEBUG
  printf("rtems-ext2fs: blockio: read: bdlock:%d size:%d block:%d count:%i\n",
         bdblock, size, block, count);
#endif
  
  cp = buf;
  while (size > 0) {
    rtems_bdbuf_buffer* bdbuf;
    rtems_status_code   sc;
    ssize_t             length;
    sc = rtems_bdbuf_read(data->dev, bdblock, &bdbuf);
    if (sc != RTEMS_SUCCESSFUL) {
      retval = EIO;
      goto error_out;
    }
    length = size > data->dd->block_size ? data->dd->block_size : size;
    memcpy(cp, bdbuf->buffer, length);
    sc = rtems_bdbuf_release(bdbuf);
    if (sc != RTEMS_SUCCESSFUL) {
      retval = EIO;
      goto error_out;
    }
    actual += length;
    size -= length;
    cp += length;
    bdblock++;
  }
  return 0;

error_out:
  if (channel->read_error)
    return (channel->read_error)(channel, block, count, buf,
                                 size, actual, retval);
  return retval;
}

static errcode_t rtems_ext2fs_blockio_read_blk(io_channel    channel,
                                               unsigned long block,
                                               int           count,
                                               void          *buf)
{
  return rtems_ext2fs_blockio_read_blk64(channel, block, count, buf);
}

static errcode_t rtems_ext2fs_blockio_write_blk64(io_channel         channel,
                                                  unsigned long long block,
                                                  int                count,
                                                  const void        *buf)
{
  struct rtems_ext2fs_blockio_private_data *data;
  const char                               *cp;
  ssize_t                                   size;
  int                                       actual = 0;
  errcode_t                                 retval = 0;
  rtems_blkdev_bnum                         bdblock;

  EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
  data = (struct rtems_ext2fs_blockio_private_data *) channel->private_data;
  EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_RTEMS_BLOCKIO_CHANNEL);

  /*
   * Handle the strange count value. If -ve the count is in bytes
   * and if +ve the count is in the block size.
   */
  size = (count < 0) ? -count : count * channel->block_size;
  
  /*
   * Map the block number from the channel block to the bdbuf block
   * number.
   */
  bdblock = (block * channel->block_size) / data->dd->block_size;

#if RTEMS_EXT2FS_BLOCKIO_DEBUG
  printf("rtems-ext2fs: blockio: write: bdlock:%d size:%d block:%d count:%d\n",
         bdblock, size, block, count);
#endif
  
  cp = buf;
  while (size > 0) {
    rtems_bdbuf_buffer* bdbuf;
    rtems_status_code   sc;
    ssize_t             length;
    sc = rtems_bdbuf_get(data->dev, bdblock, &bdbuf);
    if (sc != RTEMS_SUCCESSFUL) {
      retval = EIO;
      goto error_out;
    }
    length = size > data->dd->block_size ? data->dd->block_size : size;
    memcpy(bdbuf->buffer, cp, length);
    sc = rtems_bdbuf_release_modified(bdbuf);
    if (sc != RTEMS_SUCCESSFUL) {
      retval = EIO;
      goto error_out;
    }
    actual += length;
    size -= length;
    cp += length;
    bdblock++;
  }
  return retval;

error_out:
  if (channel->write_error)
    return (channel->write_error)(channel, block, count, buf,
                                  size, actual, retval);
  return retval;
}

static errcode_t rtems_ext2fs_blockio_write_blk(io_channel    channel,
                                                unsigned long block,
                                                int           count,
                                                const void   *buf)
{
  return rtems_ext2fs_blockio_write_blk64(channel, block, count, buf);
}

/*
 * Flush data buffers to disk.
 */
static errcode_t rtems_ext2fs_blockio_flush(io_channel channel)
{
  struct rtems_ext2fs_blockio_private_data *data;
  errcode_t                          retval = 0;
  rtems_status_code                  sc;
  
  EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
  data = (struct rtems_ext2fs_blockio_private_data *) channel->private_data;
  EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_RTEMS_BLOCKIO_CHANNEL);

  sc = rtems_bdbuf_syncdev(data->dev);
  if (sc != RTEMS_SUCCESSFUL)
    retval = EIO;
  
  return retval;
}

static errcode_t rtems_ext2fs_blockio_set_option(io_channel  channel,
                                                 const char *option,
                                                 const char *arg)
{
  struct rtems_ext2fs_blockio_private_data *data;

  EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
  data = (struct rtems_ext2fs_blockio_private_data *) channel->private_data;
  EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_RTEMS_BLOCKIO_CHANNEL);

  return EXT2_ET_INVALID_ARGUMENT;
}

dev_t rtems_ext2fs_blockio_dev(ext2_filsys fs)
{
  struct rtems_ext2fs_blockio_private_data *data;
  io_channel                                channel;

  channel = fs->image_io;
  EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
  data = (struct rtems_ext2fs_blockio_private_data *) channel->private_data;
  EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_RTEMS_BLOCKIO_CHANNEL);
  
  return data->dev;
}
