/*
 *  COPYRIGHT (c) 2009 Chris Johns <chrisj@rtems.org>
 *
 *  COPYRIGHT (c) 1989-2009 On-Line Applications Research Corporation (OAR).
 *
 *  The license and distribution terms for this file may be
 *  found in the file LICENSE in this distribution or at
 *  http://www.rtems.com/license/LICENSE.
 *
 *  $Id: memfile.c,v 1.30 2009/04/29 08:31:27 ccj Exp $
 */
/**
 * @file
 *
 * @ingroup rtems-ext2fs
 *
 * RTEMS EXT2 File Handlers
 *
 * This file contains the set of handlers used to process operations on
 * EXT2 file nodes.
 */

#if HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <errno.h>

#include <rtems.h>
#include "rtems-ext2fs-private.h"

/*
 *  rtems_ext2fs_file_open
 *
 *  This routine processes the open() system call.  Note that there is
 *  nothing special to be done at open() time.
 */

static int rtems_ext2fs_file_open(
  rtems_libio_t *iop,
  const char    *pathname,
  uint32_t       flag,
  uint32_t       mode
)
{
  ext2_filsys fs = rtems_ext2fs_dev(&iop->pathinfo);
  ext2_ino_t  ino = rtems_ext2fs_get_iop_ino(iop);
	ext2_file_t file_info;
  errcode_t   rc;
  int         ext2_flags = 0;

  if ((flag & O_WRONLY) || (flag & O_RDWR))
    ext2_flags = EXT2_FILE_CREATE | EXT2_FILE_WRITE;

#if RTEMS_EXT2FS_DEBUG_FILE_OPEN
  printf("rtems-ext2fs: file-open: path:%s ino:%d flags:%04lx mode:%04lx ext2-flags:%04x\n",
         pathname, ino, flag, mode, ext2_flags);
#endif
  
	rc = ext2fs_file_open(fs, ino, ext2_flags, &file_info);
	if (rc)
    return rtems_ext2fs_remap_error("file_open: open", rc);

#if RTEMS_EXT2FS_DEBUG_FILE_OPEN
  printf("rtems-ext2fs: file-open: file-info:%p\n", file_info);
#endif
  
  iop->file_info = file_info;
  
  return 0;
}

/*
 *  rtems_ext2fs_file_close
 *
 *  This routine processes the close() system call.  Note that there is
 *  nothing to flush at this point.
 */

static int rtems_ext2fs_file_close(
  rtems_libio_t *iop
)
{
  ext2_file_t        file_info = iop->file_info;
  struct ext2_inode* inode;
  errcode_t          rc;

#if RTEMS_EXT2FS_DEBUG_FILE_CLOSE
  printf("rtems-ext2fs: file-close: file-info:%p\n", file_info);
#endif

  ext2fs_file_get_inode(file_info, &inode);

  rc = ext2fs_write_inode(rtems_ext2fs_dev(&iop->pathinfo),
                          rtems_ext2fs_get_pathloc_ino(&iop->pathinfo),
                          inode);
  if (rc)
    return rtems_ext2fs_remap_error("file_close: close write inode", rc);

  rc = ext2fs_file_close(file_info);
  if (rc)
    return rtems_ext2fs_remap_error("file_close: close", rc);
  
  return 0;
}

/*
 *  rtems_ext2fs_file_read
 *
 *  This routine processes the read() system call.
 */

ssize_t rtems_ext2fs_file_read(
  rtems_libio_t *iop,
  void          *buffer,
  size_t         count
)
{
  ext2_file_t        file_info = iop->file_info;
  unsigned int       read;
  struct ext2_inode* inode;
  errcode_t          rc;

#if RTEMS_EXT2FS_DEBUG_FILE_READ
  printf("rtems-ext2fs: file-read: file-info:%p count:%ld\n", file_info, count);
#endif
  
  ext2fs_file_get_inode(file_info, &inode);

  rc = ext2fs_file_read(file_info, buffer, count, &read);
  if (rc)
    return rtems_ext2fs_remap_error("file_read: read", rc);

  inode->i_ctime = inode->i_atime = time(NULL);
  
  return (size_t) read;
}

/*
 *  rtems_ext2fs_file_write
 *
 *  This routine processes the write() system call.
 */

ssize_t rtems_ext2fs_file_write(
  rtems_libio_t *iop,
  const void    *buffer,
  size_t         count
)
{
  ext2_file_t        file_info = iop->file_info;
  unsigned int       write;
  rtems_off64_t      before;
  struct ext2_inode* inode;
  time_t             tm;
  errcode_t          rc;

#if RTEMS_EXT2FS_DEBUG_FILE_WRITE
  printf("rtems-ext2fs: file-write: file-info:%p count:%ld\n", file_info, count);
#endif
  
  ext2fs_file_get_inode(file_info, &inode);

  before = EXT2_I_SIZE(inode);
  
  rc = ext2fs_file_write(file_info, buffer, count, &write);
  if (rc)
    return rtems_ext2fs_remap_error("file_write: write", rc);

  tm = time(NULL);
  
  if (before != EXT2_I_SIZE(inode))
    inode->i_ctime = tm;

  inode->i_mtime = tm;
  
  return (size_t) write;
}

/*
 *  rtems_ext3fs_file_ioctl
 *
 *  This routine processes the ioctl() system call.
 *
 *  NOTE:  No ioctl()'s are currently supported for ext2fs files.
 */

int rtems_ext2fs_file_ioctl(
  rtems_libio_t *iop,
  uint32_t       command,
  void          *buffer
)
{
  return 0;
}

/*
 *  rtems_ext2fs_file_lseek
 *
 *  This routine processes the lseek() system call.
 */

rtems_off64_t rtems_ext2fs_file_lseek(
  rtems_libio_t   *iop,
  rtems_off64_t    offset,
  int              whence
)
{
  ext2_file_t  file_info = iop->file_info;
  unsigned int new_pos;
  errcode_t    rc;

#if RTEMS_EXT2FS_DEBUG_FILE_LSEEK
  printf("rtems-ext2fs: file-lseek: file-info:%p offset:%Ld\n", file_info, offset);
#endif
  
  rc = ext2fs_file_lseek(file_info, iop->offset, SEEK_SET, &new_pos);
  if (rc)
    return rtems_ext2fs_remap_error("file_lseek: lseek", rc);
  return iop->offset;
}

/*
 *  rtems_ext2fs_stat
 *
 *  This rtems_ext2fs_stat() can be used.
 */

/*
 *  rtems_ext2fs_file_ftruncate
 *
 *  This routine processes the ftruncate() system call.
 */

int rtems_ext2fs_file_ftruncate(
  rtems_libio_t        *iop,
  rtems_off64_t         length
)
{
  ext2_file_t  file_info = iop->file_info;
  errcode_t    rc;

#if RTEMS_EXT2FS_DEBUG_FILE_FTRUNC
  printf("rtems-ext2fs: file-ftrunc: file-info:%p length:%Ld\n", file_info, length);
#endif
  
  rc = ext2fs_file_set_size(file_info, length);
  if (rc)
    return rtems_ext2fs_remap_error("file_ftruncate: set size", rc);

  return 0;
}

/*
 *  rtems_ext2fs_rmnod
 *
 *  This rtems_ext2fs_rmnod can be used.
 */

/*
 *  Set of operations handlers for operations on ext2fs files.
 */

const rtems_filesystem_file_handlers_r rtems_ext2fs_file_handlers = {
  .open_h      = rtems_ext2fs_file_open,
  .close_h     = rtems_ext2fs_file_close,
  .read_h      = rtems_ext2fs_file_read,
  .write_h     = rtems_ext2fs_file_write,
  .ioctl_h     = rtems_ext2fs_file_ioctl,
  .lseek_h     = rtems_ext2fs_file_lseek,
  .fstat_h     = rtems_ext2fs_stat,
  .fchmod_h    = rtems_ext2fs_fchmod,
  .ftruncate_h = rtems_ext2fs_file_ftruncate,
  .fpathconf_h = NULL,
  .fsync_h     = rtems_ext2fs_fdatasync,
  .fdatasync_h = rtems_ext2fs_fdatasync,
  .fcntl_h     = rtems_ext2fs_fcntl,
  .rmnod_h     = rtems_ext2fs_rmnod
};
