/*
 *  $Id$
 *
 * RTEMS Project (http://www.rtems.org/)
 *
 * Copyright 2007 Chris Johns (chrisj@rtems.org)
 */

/**
 * Basic File System test. This is to just check a file system
 * down work as expected.
 *
 * Please add more tests
 */

#include <errno.h>
#include <setjmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <rtems.h>

#include "fs-test.h"

/**
 * The block size of files when creating.
 */
#define FILE_BLOCK_SIZE (256)

/**
 * The types for the file-system test.
 */
typedef enum fst_type
{
  fst_file,
  fst_dir
} fst_type;

/**
 * Jump to the top on error.
 */
static jmp_buf top;

/**
 * Stat a path returning the stat buffer.
 *
 * @param path Path to get the stat buffer of.
 * @param buf The stat buffer.
 */
static void
checked_stat (const char* path, struct stat* buf)
{
  if (stat (path, buf) < 0)
  {
    printf ("error: stat of '%s' failed: %s (%d)\n",
            path, strerror (errno), errno);
    longjmp (top, 1);
  }
}

/**
 * Return if the path points to a directory.
 *
 * @param path Path to check if a directory.
 */
static void
check_dir (const char* path)
{
  struct stat ds;
  checked_stat (path, &ds);
  if (!S_ISDIR(ds.st_mode))
  {
    printf ("error: path '%s' is not a directory\n", path);
    longjmp (top, 1);
  }
}

/**
 * Return if the change to the path is successful.
 *
 * @param path Path to change directory to.
 */
static void
change_dir (const char* path)
{
  check_dir (path);
  if (chdir (path) < 0)
  {
    printf ("error: change directory to '%s' failed: %s (%d)\n",
            path, strerror (errno), errno);
    longjmp (top, 1);
  }
}

/**
 * Return if the directory is created. The path is checked
 * to insure the full path exists and if it does not all parts
 * are made.
 *
 * @param dirpath Path to the directory to create.
 */
static bool
create_dir (const char* dirpath, bool jump_on_error)
{
  /*
   * The local copy can be written to.
   */
  char* path;
  char* next;

  /*
   * If the directory path is absolute take that as the full path
   * to create. If the path is relative get the current working
   * directory and pre-pend it to th directoy path.
   */

  if (*dirpath == '/')
  {
    path = strdup (dirpath);
    if (!path)
    {
      printf ("error: could not duplicate path in create_dir\n");
      longjmp (top, 1);
    }
  }
  else
  {
    /*
     * Find a buffer large enough for the current working directory.
     */
    size_t size = 256;
    path = NULL;
    while (path == NULL)
    {
      char* cwd = malloc (size);
      if (!cwd)
      {
        printf ("error: could not allocate memory in create_dir\n");
        longjmp (top, 1);
      }
      
      if (getcwd (cwd, size) == cwd)
      {
        path = malloc (strlen (cwd) + strlen (dirpath) + 2);

        if (!path)
        {
          free (cwd);
          printf ("error: could not duplicate path in create_dir\n");
          longjmp (top, 1);
        }

        strcpy (path, cwd);
        strcat (path, "/"); 
        strcpy (path, dirpath);

        free (cwd);
        break;
      }

      free (cwd);
      size += 256;
    }
  }
  
  next = path;

  /*
   * Break the path up into its parts and check to see if
   * each part exists. It a part does not exist create it.
   * If the part is not a directory report as an error.
   */
  
  while (next)
  {
    char separator = '/';
    struct stat ds;

    /*
     * Remove the separator and replace with a nul. If
     * no separator is found this is the end of the path.
     */
    
    next = strchr (next, '/');
    
    if (next)
    {
      separator = *next;
      *next = '\0';
    }

    if (stat (path, &ds) == 0)
    {
      if (!S_ISDIR(ds.st_mode))
      {
        printf ("error: path exists and is not a directory: %s\n", path);
        free (path);
        longjmp (top, 1);
      }
    }
    else if (errno != ENOENT)
    {
      printf ("error: path stat failure: %s: %s (%d)\n",
              path, strerror (errno), errno);
      free (path);
      longjmp (top, 1);
    }
    else
    {
      if (mkdir (path, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH) < 0)
      {
        free (path);
        if (jump_on_error)
        {
          printf ("error: make directory '%s' failed: %s (%d)\n",
                  path, strerror (errno), errno);
          longjmp (top, 1);
        }
        return false;
      }
    }

    if (next)
    {
      *next = separator;
      next++;
    }
  }

  free (path);

  return true;
}

/**
 * Return if the file is created. The path is checked
 * to insure the full path exists and if it does not all parts
 * are made. The file is fill with repeating numbers.
 *
 * @param filepath Path to the file to create.
 */
static bool
create_file (const char* filepath, size_t size, bool jump_on_error)
{
  unsigned char* block;
  unsigned char  pattern = 0;
  int fd;
  
  /*
   * Does the file path contain directories ?
   */
  if (strchr (filepath, '/') != NULL)
  {
    /*
     * Get a copy of the path and search backwards for the
     * directory separator.
     */
    
    char* dirpath = strdup (filepath);
    char* end = dirpath + strlen (dirpath);

    while (*end != '/')
      end--;

    *end = '\0';

    create_dir (dirpath, true);

    free (dirpath);
  }

  fd = open (filepath,
             O_CREAT | O_TRUNC | O_WRONLY,
             S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);

  if (fd < 0)
  {
    if (jump_on_error)
    {
      printf ("error: open failed: %s: %s (%d)\n", filepath,
              strerror (errno), errno);
      longjmp (top, 1);
    }
    return false;
  }

  block = malloc (FILE_BLOCK_SIZE);

  if (!block)
  {
    printf ("error: no memory in create_file\n");
    close (fd);
    longjmp (top, 1);
  }
  
  while (size)
  {
    size_t bsize = size < FILE_BLOCK_SIZE ? size : FILE_BLOCK_SIZE;

    memset (block, pattern++, FILE_BLOCK_SIZE);

    if (write (fd, block, bsize) < 0)
    {
      printf ("error: file write failed: %s: %s (%d)\n",
              filepath, strerror (errno), errno);
      free (block);
      close (fd);
      longjmp (top, 1);
    }

    size -= bsize;
  }

  free (block);
  
  if (close (fd) < 0)
  {
    printf ("error: file close failed: %s: %s (%d)\n",
            filepath, strerror (errno), errno);
    longjmp (top, 1);
  }

  return true;
}

/**
 * Remove directory.
 *
 * @param path The path to the directory to remove.
 */
void
remove_dir (const char* path)
{
  if (rmdir (path) < 0)
  {
    printf ("error: removing directory: %s: %s (%d)\n",
            path, strerror (errno), errno);
    longjmp (top, 1);
  }
}

/**
 * Remove file.
 *
 * @param path The path to the file to remove.
 */
void
remove_file (const char* path)
{
  if (unlink (path) < 0)
  {
    printf ("error: removing file: %s: %s (%d)\n",
            path, strerror (errno), errno);
    longjmp (top, 1);
  }
}

/**
 * Check the limit of a directory by creating directories or files
 * up to the limit.
 *
 * @param base 
 * @param type 
 * @param limit 
 */
void
check_dir_limit (const char* base, fst_type type, int limit)
{
  int count = 0;
  int c;
  
  create_dir (base, true);
  change_dir (base);
  
  while (count < limit)
  {
    char path[16];

    snprintf (path, sizeof (path), "tst%04d", count);
    
    if (type == fst_file)
      create_file (path, 64, true);
    else
      create_dir (path, true);

    count++;
  }

  for (c = 0; c < count; c++)
  {
    char path[16];

    snprintf (path, sizeof (path), "tst%04d", c);

    if (type == fst_file)
      remove_file (path);
    else
      remove_dir (path);
  }   
}

/**
 * Find the limit of a directory by creating directories or files
 * until an error is detected.
 *
 * @param base 
 * @param type 
 * @param limit 
 */
int
find_dir_limit (const char* base, fst_type type, int limit)
{
  int count = 0;
  int c;
  
  create_dir (base, true);
  change_dir (base);
  
  while (count < limit)
  {
    char path[16];

    snprintf (path, sizeof (path), "tst%04d", count);
    
    if (type == fst_file)
    {
      if (!create_file (path, 64, false))
        break;
    }
    else
    {
      if (!create_dir (path, false))
        break;
    }

    count++;
  }

  for (c = 0; c < count; c++)
  {
    char path[16];

    snprintf (path, sizeof (path), "tst%04d", c);

    if (type == fst_file)
      remove_file (path);
    else
      remove_dir (path);
  }

  return count;
}

/**
 * Set the file size then stat to make sure the size is what we asked for.
 */
bool
set_size_then_check (int fd, int pos, const char* what)
{
  struct stat sb;

  if (ftruncate (fd, pos) < 0)
  {
    printf ("error: ftruncate %s failed: pos: %d: %s (%d)\n",
            what, pos, strerror (errno), errno);
    return false;
  }

  if (fstat (fd, &sb) < 0)
  {
    printf ("error: stat %s failed: %s (%d)\n", what, strerror (errno), errno);
    return false;
  }

  if (sb.st_size != pos)
  {
    printf ("error: ftruncate %s invalid: file size not %d; is %Lu\n",
            what, pos, sb.st_size);
    return false;
  }

  return true;
}

bool
read_words_and_verify (const char* what, int fd, size_t count, off_t offset)
{
  uint8_t buffer[sizeof (uint16_t)];
  int     i;
  
  if (lseek (fd, offset, SEEK_SET) < 0)
  {
    printf ("error: lseeking for %s failed: %s (%d)\n",
            what, strerror (errno), errno);
    return false;
  }

  for (i = 0; i < (count / 2); i++)
  {
    int rc = read (fd, buffer, sizeof (uint16_t));
    if (rc != sizeof (uint16_t))
    {
      printf ("error: read pattern for %s failed: read at %u: %d: %s (%d)\n",
              what, (unsigned int) (i + offset), rc, strerror (errno), errno);
      return false;
    }
    
    if ((buffer[0] != (i >> 8)) || buffer[1] != (i & 0xff))
    {
      printf ("error: read pattern match for %s failed: : read: %02x%02x, wanted: %04x\n",
              what, buffer[0], buffer[1], i);
      return false;
    }
  }

  return true;
}

/**
 * Create an empty file then stat it. It should be 0. Then lseek to a location
 * and stat again. It should still be 0. Read from the file and it should be
 * EOF. Write to the file then stat and check the file. Seek to the start of
 * the file and read the data up to the first seek location. It should be 0.
 * Seek to the original position and then extend the file using
 * ftruncate. Check the stat is the correct file. Read the block of data
 * originally written has been read back. The ftruncate call should not move
 * the file pointer. Perform a range or ftruncate with different values to make
 * sure the file systems handle block alignment issues.
 *
 * @param base 
 */
#define LSEEK_BUF_SIZE (512)
#define LSEEK_POS_1    (1024)
bool
check_lseek_ftruncate_read_write (const char* base)
{
  const char* name = "tstlseekftruncate";
  uint8_t     buffer[LSEEK_BUF_SIZE];
  uint32_t    block_size;
  struct stat sb;
  int         pos;
  int         count;
  int         fd;
  bool        result = true;
  int         rc;
  int         i;

  create_dir (base, true);
  change_dir (base);
  
  fd = open (name,
             O_CREAT | O_TRUNC | O_RDWR,
             S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);

  if (fd < 0)
  {
    printf ("error: open failed: %s: %s (%d)\n",
            name, strerror (errno), errno);
    return false;
  }

  if (fstat (fd, &sb) < 0)
  {
    printf ("error: stat failed: %s: %s (%d)\n",
            name, strerror (errno), errno);
    result = false;
    goto err;
  }

  block_size = sb.st_blksize;
  
  if (sb.st_size != 0)
  {
    printf ("error: create size invalid: file size not 0; is %Lu\n", sb.st_size);
    result = false;
    goto err;
  }

  if (read (fd, buffer, LSEEK_BUF_SIZE) < 0)
  {
    printf ("error: read of empty file failed: %s (%d)\n",
            strerror (errno), errno);
    result = false;
    goto err;
  }
  
  if (lseek (fd, LSEEK_POS_1, SEEK_SET) < 0)
  {
    printf ("error: first lseek failed: pos: %d: %s (%d)\n",
            LSEEK_POS_1, strerror (errno), errno);
    result = false;
    goto err;
  }

  if (fstat (fd, &sb) < 0)
  {
    printf ("error: stat failed: %s (%d)\n", strerror (errno), errno);
    result = false;
    goto err;
  }

  if (sb.st_size != 0)
  {
    printf ("error: lseek 1 size invalid: file size not 0; is %Lu\n", sb.st_size);
    result = false;
    goto err;
  }

  rc = read (fd, buffer, LSEEK_BUF_SIZE); 
  if (rc != 0)
  {
    printf ("error: read of empty file after seek failed: read: %d: %s (%d)\n",
            rc, strerror (errno), errno);
    result = false;
    goto err;
  }
  
  for (i = 0; i < LSEEK_BUF_SIZE; i++)
    buffer[i] = i & 0xff;

  rc = write (fd, buffer, LSEEK_BUF_SIZE); 
  if (rc != LSEEK_BUF_SIZE)
  {
    printf ("error: write empty file after seek has failed: write: %d: %s (%d)\n",
            rc, strerror (errno), errno);
    result = false;
    goto err;
  }
  
  if (fstat (fd, &sb) < 0)
  {
    printf ("error: stat after write failed: %s (%d)\n", strerror (errno), errno);
    result = false;
    goto err;
  }

  if (sb.st_size != (LSEEK_POS_1 + LSEEK_BUF_SIZE))
  {
    printf ("error: post write size wrong: file size not %d; is %Lu\n",
            LSEEK_POS_1 + LSEEK_BUF_SIZE, sb.st_size);
    result = false;
    goto err;
  }

  if (lseek (fd, 0, SEEK_SET) < 0)
  {
    printf ("error: lseek to pos 0 failed: %s (%d)\n", strerror (errno), errno);
    result = false;
    goto err;
  }

  count = LSEEK_POS_1;
  while (count)
  {
    int length = LSEEK_BUF_SIZE;
    
    if (count < length)
      length = count;
    
    rc = read (fd, buffer, length);    
    if (rc != length)
    {
      printf ("error: read lseek space failed: read: %d: %s (%d)\n",
              rc, strerror (errno), errno);
      result = false;
      goto err;
    }

    for (i = 0; i < length; i++)
    {
      if (buffer[i] != 0)
      {
        printf ("error: read lseek space failed: not zero at %d -> %02x \n",
                i, buffer[i]);
        result = false;
        goto err;
      }
    }

    count -= length;
  }

  count = LSEEK_BUF_SIZE;
  while (count)
  {
    int length = LSEEK_BUF_SIZE;
    
    if (count < length)
      length = count;
    
    rc = read (fd, buffer, length);
    if (rc != length)
    {
      printf ("error: read write space failed: read: %d: %s (%d)\n",
              rc, strerror (errno), errno);
      result = false;
      goto err;
    }

    for (i = 0; i < length; i++)
    {
      if (buffer[i] != ((uint8_t) (i & 0xff)))
      {
        printf ("error: read write space failed: not zero at %d -> %02x \n",
                i, buffer[i]);
        result = false;
        goto err;
      }
    }

    count -= length;
  }

  /*
   * Play with the file system's block boundaries.
   */
  
  pos = block_size * 8;
  if (!set_size_then_check (fd, pos, "8 x blksize"))
  {
    result = false;
    goto err;
  }

  if (!set_size_then_check (fd, pos - 1, "8 x blksize"))
  {
    result = false;
    goto err;
  }

  if (!set_size_then_check (fd, pos + 1, "8 x blksize"))
  {
    result = false;
    goto err;
  }
  
  srand ((unsigned int) time (NULL));
  
  for (i = 0; i < 100; i++)
  {
    pos = LSEEK_POS_1 + (rand () % (block_size * 8));

    if (!set_size_then_check (fd, pos + 1, "random pos"))
    {
      result = false;
      goto err;
    }
  }

  if (ftruncate (fd, 0) < 0)
  {
    printf ("error: ftruncate to 0 failed: pos: 0: %s (%d)\n",
            strerror (errno), errno);
    result = false;
    goto err;
  }

  if (lseek (fd, 0, SEEK_SET) < 0)
  {
    printf ("error: lseek to pos 0 failed: %s (%d)\n", strerror (errno), errno);
    result = false;
    goto err;
  }

  for (i = 0; i < (LSEEK_BUF_SIZE / sizeof (uint16_t)); i++)
  {
    buffer[(i * 2) + 0] = (uint8_t) (i >> 8);
    buffer[(i * 2) + 1] = (uint8_t) i;
  }

  rc = write (fd, buffer, LSEEK_BUF_SIZE); 
  if (rc != LSEEK_BUF_SIZE)
  {
    printf ("error: write pattern block has failed: write: %d: %s (%d)\n",
            rc, strerror (errno), errno);
    result = false;
    goto err;
  }

  if (!read_words_and_verify ("block write", fd, LSEEK_BUF_SIZE, 0))
  {
    result = false;
    goto err;
  }

  if (ftruncate (fd, 0) < 0)
  {
    printf ("error: ftruncate to 0 failed: pos: 0: %s (%d)\n",
            strerror (errno), errno);
    result = false;
    goto err;
  }

  if (lseek (fd, 0, SEEK_SET) < 0)
  {
    printf ("error: lseek to pos 0 failed: %s (%d)\n", strerror (errno), errno);
    result = false;
    goto err;
  }

  for (i = 0; i < (LSEEK_BUF_SIZE / sizeof (uint16_t)); i++)
  {
    buffer[0] = (uint8_t) (i >> 8);
    buffer[1] = (uint8_t) i;

    rc = write (fd, buffer, sizeof (uint16_t)); 
    if (rc != sizeof (uint16_t))
    {
      printf ("error: write pattern words has failed: write: %d: %s (%d)\n",
              rc, strerror (errno), errno);
      result = false;
      goto err;
    }
  }

  if (!read_words_and_verify ("block word", fd, LSEEK_BUF_SIZE, 0))
  {
    result = false;
    goto err;
  }
  
err:
  close (fd);
  remove_file (name);
  return result;
}

/**
 * Test the file system.
 */
void
filesystem_test (const char* base)
{
  int limit;
  
  printf ("File System test: %s\n", base);

/*
   * Catch any error and return.
   */
  if (setjmp (top))
    return;

  /*
   * Find out how many files will fit in the root directory.
   */
  printf ("Find root directory limits: files\n");
  limit = find_dir_limit (base, fst_file, 1000);
  printf ("Root directory limits: %d\n", limit);
  
  /*
   * Make sure we can fill a subdirectory with more.
   */
  printf ("Checking directory limits: files\n");
  check_dir_limit (base, fst_file, limit);

  /*
   * Repeat with directories.
   */
  printf ("Checking directory limits: dir\n");
  check_dir_limit (base, fst_dir, limit);

  /*
   * Check lseek and ftruncate work.
   */
  printf ("Checking lseek, ftruncate, read, and write\n");
  if (!check_lseek_ftruncate_read_write (base))
    return;
  
  printf ("File System test: passed\n");
}
