/*
 *  $Id$
 *
 * RTEMS Project (http://www.rtems.org/)
 *
 * Copyright 2007 Chris Johns (chrisj@rtems.org)
 */

/**
 * File-system block driver test program.
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>

#include <rtems.h>
#include <rtems/bdbuf.h>
#include <rtems/blkdev.h>
#include <rtems/diskdevs.h>
#include <rtems/dosfs.h>
#include <rtems/error.h>
#include <rtems/nvdisk.h>
#include <rtems/fsmount.h>
#include <rtems/ramdisk.h>
#include <rtems/shell.h>
#include <rtems/untar.h>

#include <rtems/rtems-rfs.h>
#include <rtems/rtems-rfs-format.h>
#include <rtems/rtems-rfs-shell.h>

#include <rtems/nvdisk-sram.h>
#include <ext2fs/rtems-ext2fs.h>

#if RTEMS_APP_FLASHDISK
#include <rtems/flashdisk.h>
#include <libchip/am29lv160.h>
#endif

#if RTEMS_APP_IDEDISK
#include <libchip/ide_ctrl.h>
#include <libchip/ata.h>
#endif

#if pc586
#include "pc386-gdb.h"
int remote_debug;
#endif

#include "fs-test.h"

/*
 *  The tarfile is built automatically externally so we need to account
 *  for the leading symbol on the names.
 */
#if defined(__sh__)
  #define SYM(_x) _x
#else
  #define SYM(_x) _ ## _x
#endif

/**
 * The shell-init tarfile parameters.
 */
extern int SYM(binary_shell_init_tar_start);
extern int SYM(binary_shell_init_tar_size);

#define TARFILE_START SYM(binary_shell_init_tar_start)
#define TARFILE_SIZE  SYM(binary_shell_init_tar_size)

/**
 * The RAM Disk configuration.
 */
rtems_ramdisk_config rtems_ramdisk_configuration[] =
{
  {
    block_size: 512,
    block_num:  1984 * 2,
    location:   NULL
  }
};

/**
 * The number of RAM Disk configurations.
 */
size_t rtems_ramdisk_configuration_size = 1;

/**
 * Create the RAM Disk Driver entry.
 */
rtems_driver_address_table rtems_ramdisk_io_ops = {
  initialization_entry: ramdisk_initialize,
  open_entry:           rtems_blkdev_generic_open,
  close_entry:          rtems_blkdev_generic_close,
  read_entry:           rtems_blkdev_generic_read,
  write_entry:          rtems_blkdev_generic_write,
  control_entry:        rtems_blkdev_generic_ioctl
};

#if RTEMS_APP_FLASHDISK
/**
 * The Flash Device Driver configuration.
 */
const rtems_am29lv160_config rtems_am29lv160_configuration[] =
{
  {
    bus_8bit: 0,
    base:     (void*) 0xFFE00000
  }
};

/**
 * The number of AM29LV160 configurations.
 */
uint32_t rtems_am29lv160_configuration_size = 1;

/**
 * The Flash Segment descriptor. The mcf5235 board uses the
 * bottom part of the flash for the dBug monitor.
 */
const rtems_fdisk_segment_desc rtems_mcf5235_segment_descriptor[] =
{
  {
    count:   26,
    segment: 0,
    offset:  0x00050000,
    size:    RTEMS_FDISK_KBYTES (64)
  }
};

/**
 * The Flash Device descriptor.
 */
const rtems_fdisk_device_desc rtems_mcf5235_device_descriptor[] =
{
  {
    segment_count: 1,
    segments:      &rtems_mcf5235_segment_descriptor[0],
    flash_ops:     &rtems_am29lv160_handlers
  }
};

/**
 * The Flash Disk configuration.
 */
const rtems_flashdisk_config rtems_flashdisk_configuration[] =
{
  {
    block_size:         512,
    device_count:       1,
    devices:            &rtems_mcf5235_device_descriptor[0],
    flags:              RTEMS_FDISK_BLANK_CHECK_BEFORE_WRITE,
    unavail_blocks:     256,
    compact_segs:       100,
    avail_compact_segs: 100,
    info_level:         0
  }
};

/**
 * The number of Flash Disk configurations.
 */
size_t rtems_flashdisk_configuration_size = 1;

/**
 * Create the Flash Disk Driver entry.
 */
rtems_driver_address_table rtems_flashdisk_io_ops = {
  initialization_entry: rtems_fdisk_initialize,
  open_entry:           rtems_blkdev_generic_open,
  close_entry:          rtems_blkdev_generic_close,
  read_entry:           rtems_blkdev_generic_read,
  write_entry:          rtems_blkdev_generic_write,
  control_entry:        rtems_blkdev_generic_ioctl
};
#endif

/**
 * The NV Device descriptor. For this test it is just DRAM.
 */
rtems_nvdisk_device_desc rtems_nv_heap_device_descriptor[] =
{
  {
    flags:  0,
    base:   0,
    size:   2 * 1024 * 1024,
    nv_ops: &rtems_nvdisk_sram_handlers
  }
};

/**
 * The NV Disk configuration.
 */
const rtems_nvdisk_config rtems_nvdisk_configuration[] =
{
  {
    block_size:         512,
    device_count:       1,
    devices:            &rtems_nv_heap_device_descriptor[0],
    flags:              0,
    info_level:         0
  }
};

/**
 * The number of NV Disk configurations.
 */
size_t rtems_nvdisk_configuration_size = 1;

/**
 * Create the NV Disk Driver entry.
 */
rtems_driver_address_table rtems_nvdisk_io_ops = {
  initialization_entry: rtems_nvdisk_initialize,
  open_entry:           rtems_blkdev_generic_open,
  close_entry:          rtems_blkdev_generic_close,
  read_entry:           rtems_blkdev_generic_read,
  write_entry:          rtems_blkdev_generic_write,
  control_entry:        rtems_blkdev_generic_ioctl
};

#if RTEMS_APP_IDEDISK
/**
 * Create the IDE Disk Driver entry.
 */
rtems_driver_address_table rtems_idedisk_io_ops = {
  initialization_entry: rtems_ata_initialize,
  open_entry:           rtems_blkdev_generic_open,
  close_entry:          rtems_blkdev_generic_close,
  read_entry:           rtems_blkdev_generic_read,
  write_entry:          rtems_blkdev_generic_write,
  control_entry:        rtems_blkdev_generic_ioctl
};
#endif

/**
 * Let the IO system allocation the next available major number.
 */
#define RTEMS_DRIVER_AUTO_MAJOR (0)

/**
 * Start the RTEMS Shell.
 */
void
shell_start (void)
{
  rtems_status_code sc;
  printf ("Starting shell....\n\n");
  sc = rtems_shell_init ("fstst", 60 * 1024, 150, "/dev/console", 0, 1, NULL);
  if (sc != RTEMS_SUCCESSFUL)
    printf ("error: starting shell: %s (%d)\n", rtems_status_text (sc), sc);
}

/**
 * Run the /shell-init script.
 */
void
shell_init_script (void)
{
  rtems_status_code sc;
  printf ("Running /shell-init....\n\n");
  sc = rtems_shell_script ("fstst", 60 * 1024, 160, "/shell-init", "stdout",
                           0, 1, 1);
  if (sc != RTEMS_SUCCESSFUL)
    printf ("error: running shell script: %s (%d)\n", rtems_status_text (sc), sc);
}

int
setup_ramdisk (const char* mntpath)
{
  rtems_device_major_number major;
  rtems_status_code         sc;
  
  /*
   * Register the RAM Disk driver.
   */
  printf ("Register RAM Disk Driver: ");
  sc = rtems_io_register_driver (RTEMS_DRIVER_AUTO_MAJOR,
                                 &rtems_ramdisk_io_ops,
                                 &major);
  if (sc != RTEMS_SUCCESSFUL)
  {
    printf ("error: ramdisk driver not initialised: %s\n",
            rtems_status_text (sc));
    return 1;
  }
  
  printf ("successful\n");

  return 0;
}

int
setup_flashdisk (const char* mntpath)
{
#if RTEMS_APP_FLASHDISK
  rtems_device_major_number major;
  rtems_status_code         sc;
  
  /*
   * Register the Flash Disk driver.
   */
  printf ("Register Flash Disk Driver: ");
  sc = rtems_io_register_driver (RTEMS_DRIVER_AUTO_MAJOR,
                                 &rtems_flashdisk_io_ops,
                                 &major);
  if (sc != RTEMS_SUCCESSFUL)
  {
    printf ("error: flashdisk driver not initialised: %s\n",
            rtems_status_text (sc));
    return 1;
  }
  
  printf ("successful\n");
#endif
  return 0;
}

int
setup_nvdisk (const char* mntpath)
{
  rtems_device_major_number major;
  rtems_status_code         sc;

  /*
   * For our test we do not have any static RAM or EEPROM devices so
   * we allocate the memory from the heap.
   */
  rtems_nv_heap_device_descriptor[0].base =
    (uint32_t) malloc (rtems_nv_heap_device_descriptor[0].size);

  if (!rtems_nv_heap_device_descriptor[0].base)
  {
    printf ("error: no memory for NV disk\n");
    return 1;
  }
  
  /*
   * Register the NV Disk driver.
   */
  printf ("Register NV Disk Driver: ");
  sc = rtems_io_register_driver (RTEMS_DRIVER_AUTO_MAJOR,
                                 &rtems_nvdisk_io_ops,
                                 &major);
  if (sc != RTEMS_SUCCESSFUL)
  {
    printf ("error: nvdisk driver not initialised: %s\n",
            rtems_status_text (sc));
    return 1;
  }
  
  printf ("successful\n");

  return 0;
}

#if RTEMS_APP_IDEDISK
extern rtems_status_code rtems_ide_part_table_initialize (const char* );
#endif

int
setup_idedisk (const char* path)
{
#if RTEMS_APP_IDEDISK
  rtems_status_code sc;

  /*
   * Register the IDE Disk driver.
   */
  printf ("Read IDE Disk Partition Table: ");
  sc = rtems_ide_part_table_initialize (path);
  if (sc != RTEMS_SUCCESSFUL)
    printf ("error: ide partition table not found: %s\n",
            rtems_status_text (sc));
  else
    printf ("successful\n");

#endif
  return 0;
}

int
setup_rootfs (void)
{
  rtems_status_code sc;
  
  printf("Loading filesystem: ");
  
  sc = Untar_FromMemory((void *)(&TARFILE_START), (size_t)&TARFILE_SIZE);

  if (sc != RTEMS_SUCCESSFUL)
  {
    printf ("error: untar failed: %s\n", rtems_status_text (sc));
    return 1;
  }
  
  printf ("successful\n");

  return 0;
}

int
shell_filesystem_test (int argc, char* argv[])
{
  if (argc != 2)
  {
    printf ("error: invalid number of options\n");
    return 1;
  }

  filesystem_test (argv[1]);

  return 0;
}

int
shell_filesystem_script (int argc, char* argv[])
{
  rtems_status_code sc;

  if (argc != 2)
  {
    printf ("error: invalid number of options\n");
    return 1;
  }

  printf ("Running script: %s\n", argv[1]);
  
  sc = rtems_shell_script ("fstst", 60 * 1024, 20, argv[1], "stdout",
                           0, 1, 1);
  if (sc != RTEMS_SUCCESSFUL)
  {
    printf ("error: running shell script: %s (%d)\n",
            rtems_status_text (sc), sc);
    return 1;
  }

  return 0;
}

int
shell_fdisk_monitor (int argc, char* argv[])
{
#if RTEMS_APP_FLASHDISK
  rtems_fdisk_monitor_data data;
  const char*              driver = 0;
  uint32_t                 pages_available;
  uint32_t                 active_used_bad;
  int                      show_debug = 0;
  int                      arg;
  
  for (arg = 1; arg < argc; arg++)
  {
    if (argv[arg][0] == '-')
    {
      switch (argv[arg][1])
      {
        case 'd':
          show_debug = 1;
          break;

        default:
          printf ("error: invalid option: %s\n", argv[arg]);
          return 1;
      }
    }
    else
    {
      if (!driver)
        driver = argv[arg];
      else
      {
        printf ("error: only one driver name allowed: %s\n", argv[arg]);
        return 1;
      }
    }
  }
  
  int fd = open (driver, O_WRONLY, 0);
  if (fd < 0)
  {
    printf ("error: driver open failed: %s\n", strerror (errno));
    return 1;
  }
  
  if (ioctl (fd, RTEMS_FDISK_IOCTL_MONITORING, &data) < 0)
  {
    printf ("error: driver get monitoring data failed: %s\n", strerror (errno));
    return 1;
  }
  
  active_used_bad = data.pages_active + data.pages_used + data.pages_bad;
  pages_available = data.page_count - active_used_bad;

  printf ("Flash Disk Driver Monitoring Data: %s\n", driver);
  printf ("Capacity\t%ld bytes (Total %ld bytes)\n",
          data.block_size * (data.block_count - data.unavail_blocks),
          data.block_size * data.block_count);
  printf ("Usage\t\t%ld%% %ld bytes\n",
          (data.blocks_used * 100) / data.block_count,
          data.blocks_used * data.block_size);
  printf ("Available\t%ld%%\n", (pages_available * 100) / data.page_count);
  printf ("Efficency\t%ld%%\n",
          active_used_bad ? (data.pages_active * 100) / active_used_bad : 0);
  printf ("Active\t\t%ld%%\n",
          pages_available ? (data.pages_active * 100) / pages_available: 0);
  printf ("Used\t\t%ld%%\n",
          pages_available ? (data.pages_used * 100) / pages_available : 0);
  printf ("Bad\t\t%ld%%\n",
          pages_available ? (data.pages_bad * 100) / pages_available : 0);
  printf ("Block Size\t%ld bytes\n", data.block_size);
  printf ("Block Count\t%ld\n", data.block_count);
  printf ("Segment Count\t%ld\n", data.segment_count);
  printf ("Segments Avail\t%ld\n", data.segs_available);
  printf ("Segments Used\t%ld\n", data.segs_used);
  printf ("Segments Failed\t%ld\n", data.segs_failed);
  printf ("Segments Erases\t%ld\n", data.seg_erases);
  printf ("Page Descs\t%ld\n", data.pages_desc);
  printf ("Pages Active\t%ld\n", data.pages_active);
  printf ("Pages Used\t%ld\n", data.pages_used);
  printf ("Pages Bad\t%ld\n", data.pages_bad);
  printf ("Info Level\t%ld\n", data.info_level);

  printf ("\n");

  if (show_debug)
  {
    if (ioctl (fd, RTEMS_FDISK_IOCTL_PRINT_STATUS) < 0)
    {
      printf ("error: driver print status failed: %s\n", strerror (errno));
      return 1;
    }
  }
  
  close (fd);
#endif
  return 0;
}

int
shell_fdisk_trace (int argc, char* argv[])
{
#if RTEMS_APP_FLASHDISK
  const char* driver;
  int         level;

  if (argc != 3)
  {
    printf ("error: invalid number of options\n");
    return 1;
  }

  driver = argv[1];
  level  = strtoul (argv[2], 0, 0);
  
  int fd = open (driver, O_WRONLY, 0);
  if (fd < 0)
  {
    printf ("error: driver open failed: %s\n", strerror (errno));
    return 1;
  }
  
  if (ioctl (fd, RTEMS_FDISK_IOCTL_INFO_LEVEL, level) < 0)
  {
    printf ("error: driver set level failed: %s\n", strerror (errno));
    return 1;
  }
  
  close (fd);
#endif
  return 0;
}

int
shell_flash_erase (int argc, char* argv[])
{
#if RTEMS_APP_FLASHDISK
  const char* driver = NULL;
  int         arg;
  int         fd;
  
  for (arg = 1; arg < argc; arg++)
  {
    if (argv[arg][0] == '-')
    {
      printf ("error: invalid option: %s\n", argv[arg]);
      return 1;
    }
    else
    {
      if (!driver)
        driver = argv[arg];
      else
      {
        printf ("error: only one driver name allowed: %s\n", argv[arg]);
        return 1;
      }
    }
  }
  
  printf ("erase flash disk: %s\n", driver);
  
  fd = open (driver, O_WRONLY, 0);
  if (fd < 0)
  {
    printf ("error: flash driver open failed: %s\n", strerror (errno));
    return 1;
  }
  
  if (ioctl (fd, RTEMS_FDISK_IOCTL_ERASE_DISK) < 0)
  {
    printf ("error: flash driver erase failed: %s\n", strerror (errno));
    return 1;
  }
  
  close (fd);
  
  printf ("flash disk erased successful\n");
#endif
  return 0;
}

int
shell_fs_footprint (int argc, char* argv[])
{
#if 0
  /* Table of disk devices having the same major number */
  struct disk_device_table {
    rtems_disk_device **minor; /* minor-indexed disk device table */
    int size;            /* Number of entries in the table */
  };
  
  rtems_bdbuf_config* bdbuf_config = &rtems_bdbuf_configuration[0];
  size_t              bdbuf_usage = 0;
  size_t              diskdevs_usage = 0;
  
  /* bdbuf_initialize_pool bdbuf.c:694 calloc */
  bdbuf_usage += bdbuf_config->num * sizeof (rtems_bdbuf_buffer);
  /* bdbuf_initialize_pool bdbuf.c:703 malloc */
  bdbuf_usage += bdbuf_config->num * bdbuf_config->size;
  /* bdbuf_initialize_pool bdbuf.c:726 sema create */
  /* rtems_bdbuf_init bdbuf.c:792 calloc */
  bdbuf_usage += rtems_bdbuf_configuration_size * sizeof (rtems_bdbuf_pool);
  /* rtems_bdbuf_init bdbuf.c:818 sema create */
  /* rtems_bdbuf_init bdbuf.c:830 task create */

  /* create_disk_entry diskdevs.c:84 realloc */
  diskdevs_usage += sizeof (struct disk_device_table);
  /* create_disk_entry diskdevs.c:111 realloc */
  diskdevs_usage += sizeof (struct rtems_disk_device*);
  /* create_disk_entry diskdevs.c:124 calloc */
  diskdevs_usage += sizeof (struct rtems_disk_device);
  /* create_disk diskdevs.c:192 malloc */
  diskdevs_usage += 32;
  /* rtems_disk_io_initialize diskdevs.c:565 calloc IGNORE due to diskdevs.c:84 */
  /* rtems_disk_io_initialize diskdevs.c:570 sema create */

  printf ("BD Buf Pool size\t%ld\n", sizeof (rtems_bdbuf_pool));
  printf ("BD Buf size\t\t%ld\n", sizeof (rtems_bdbuf_buffer));
  printf ("Total\t\t%ld\n", bdbuf_usage + diskdevs_usage);
  printf ("BD Buffer\t%ld\n", bdbuf_usage);
  printf ("Disk Devices\t%ld\n", diskdevs_usage);
#endif
  return 0;
}

int
shell_nvdisk_trace (int argc, char* argv[])
{
  const char* driver;
  int         level;

  if (argc != 3)
  {
    printf ("error: invalid number of options\n");
    return 1;
  }

  driver = argv[1];
  level  = strtoul (argv[2], 0, 0);
  
  int fd = open (driver, O_WRONLY, 0);
  if (fd < 0)
  {
    printf ("error: driver open failed: %s\n", strerror (errno));
    return 1;
  }
  
  if (ioctl (fd, RTEMS_NVDISK_IOCTL_INFO_LEVEL, level) < 0)
  {
    printf ("error: driver set level failed: %s\n", strerror (errno));
    return 1;
  }
  
  close (fd);
  
  return 0;
}

int
shell_nvdisk_erase (int argc, char* argv[])
{
  const char* driver = NULL;
  int         arg;
  int         fd;
  
  for (arg = 1; arg < argc; arg++)
  {
    if (argv[arg][0] == '-')
    {
      printf ("error: invalid option: %s\n", argv[arg]);
      return 1;
    }
    else
    {
      if (!driver)
        driver = argv[arg];
      else
      {
        printf ("error: only one driver name allowed: %s\n", argv[arg]);
        return 1;
      }
    }
  }
  
  printf ("erase nv disk: %s\n", driver);
  
  fd = open (driver, O_WRONLY, 0);
  if (fd < 0)
  {
    printf ("error: nvdisk driver open failed: %s\n", strerror (errno));
    return 1;
  }
  
  if (ioctl (fd, RTEMS_NVDISK_IOCTL_ERASE_DISK) < 0)
  {
    printf ("error: nvdisk driver erase failed: %s\n", strerror (errno));
    return 1;
  }
  
  close (fd);
  
  printf ("nvdisk erased successful\n");

  return 0;
}

int
shell_bdbuf_trace (int argc, char* argv[])
{
#if RTEMS_BDBUF_TRACE
  extern bool rtems_bdbuf_tracer;
  rtems_bdbuf_tracer = !rtems_bdbuf_tracer;
  printf ("bdbuf trace: %d\n", rtems_bdbuf_tracer);
#else
  printf ("bdbuf trace disabled. Rebuild with enabled.\n");
#endif
  return 0;
}

int
disk_test_set_block_size (dev_t dev, size_t size)
{
  rtems_disk_device* dd;
  int                rc;
  
  dd = rtems_disk_obtain (dev);
  if (!dd)
  {
    printf ("error: cannot obtain disk\n");
    return 1;
  }
  
  rc = dd->ioctl (dd, RTEMS_BLKIO_SETBLKSIZE, &size);

  rtems_disk_release (dd);

  return rc;
}

int
disk_test_write_blocks (dev_t dev, int start, int count, size_t size)
{
  int                 block;
  uint32_t*           ip;
  uint32_t            value = 0;
  int                 i;
  rtems_bdbuf_buffer* bd;
  rtems_status_code   sc;
  
  if (disk_test_set_block_size (dev, size) < 0)
  {
    printf ("error: set block size failed: %s\n", strerror (errno));
    return 1;
  }

  for (block = start; block < (start + count); block++)
  {
    sc = rtems_bdbuf_read (dev, block, &bd);
    if (sc != RTEMS_SUCCESSFUL)
    {
      printf ("error: get block %d bd failed: %s\n",
              block, rtems_status_text (sc));
      return 1;
    }

    ip = (uint32_t*) bd->buffer;
    for (i = 0; i < (size / sizeof (uint32_t)); i++, ip++, value++)
      *ip = (size << 16) | value;

    sc = rtems_bdbuf_release_modified (bd);
    if (sc != RTEMS_SUCCESSFUL)
    {
      printf ("error: release block %d bd failed: %s\n",
              block, rtems_status_text (sc));
      return 1;
    }
  }

  return 0;
}

int
disk_test_block_sizes (int argc, char *argv[])
{
  struct stat st;
  char*       name;
  int         start;
  int         count;
  int         size;
  
  if (argc != (4 + 1))
  {
    printf ("error: need to supply a device path, start, block and size\n");
    return 1;
  }

  name = argv[1];
  
  if (stat (name, &st) < 0)
  {
    printf ("error: stat '%s' failed: %s\n", name, strerror (errno));
    return 1;
  }

  start = strtoul (argv[2], 0, 0);
  count = strtoul (argv[3], 0, 0);
  size  = strtoul (argv[4], 0, 0);
  
  return disk_test_write_blocks (st.st_rdev, start, count, size);
}

int rtems_shell_main_dosfsck (int argc, char *argv[]);
int rtems_ext2fs_mkfsck_main (int argc, char *argv[]);
void rtems_rfs_bitmap_unit_test (void);
int rtems_shell_main_ln(int argc, char *argv[]);
int rtems_shell_main_mknod(int argc, char *argv[]);
int test_main(void);

int
main (int argc, char* argv[])
{
  struct termios term;
  int            ret;

#if pc586
  int arg;
  for (arg = 1; arg < argc; arg++)
    if (strcmp (argv[arg], "--gdb") == 0)
      pc386_gdb_init ();
#endif

  if (tcgetattr(fileno(stdout), &term) < 0)
    printf ("error: cannot get terminal attributes: %s\n", strerror (errno));
  cfsetispeed (&term, B115200);
  cfsetospeed (&term, B115200);
  if (tcsetattr (fileno(stdout), TCSADRAIN, &term) < 0)
    printf ("error: cannot set terminal attributes: %s\n", strerror (errno));
  
  if (tcgetattr(fileno(stdin), &term) < 0)
    printf ("error: cannot get terminal attributes: %s\n", strerror (errno));
  cfsetispeed (&term, B115200);
  cfsetospeed (&term, B115200);
  if (tcsetattr (fileno(stdin), TCSADRAIN, &term) < 0)
    printf ("error: cannot set terminal attributes: %s\n", strerror (errno));
  
  printf ("\nRTEMS MS-DOS Filesystem Block Driver Test, Version " \
          PACKAGE_VERSION "\n\n");

  ret = setup_ramdisk ("/mnt/ramdisk");
  if (ret)
    exit (ret);
  
  ret = setup_flashdisk ("/mnt/flashdisk");
  if (ret)
    exit (ret);

  ret = setup_nvdisk ("/mnt/nvdisk");
  if (ret)
    exit (ret);
  
  ret = setup_idedisk ("/dev/hda");
  if (ret)
    exit (ret);
  
  ret = setup_idedisk ("/dev/hdb");
  if (ret)
    exit (ret);

  ret = setup_rootfs ();
  if (ret)
    exit (ret);

  rtems_shell_add_cmd ("fderase", "misc",
                       "fderase driver", shell_flash_erase);
  rtems_shell_add_cmd ("fdtrace", "misc",
                       "fdtrace driver level", shell_fdisk_trace);
  rtems_shell_add_cmd ("fdmon",   "misc",
                       "fdmon driver", shell_fdisk_monitor);
  rtems_shell_add_cmd ("fstst",   "misc",
                       "fstst path", shell_filesystem_test);
  rtems_shell_add_cmd ("fssize",  "misc",
                       "fssize", shell_fs_footprint);
  rtems_shell_add_cmd ("shell",  "misc",
                       "shell script", shell_filesystem_script);
  rtems_shell_add_cmd ("nverase", "misc",
                       "nverase driver", shell_nvdisk_erase);
  rtems_shell_add_cmd ("nvtrace", "misc",
                       "nvtrace driver level", shell_nvdisk_trace);
  rtems_shell_add_cmd ("dosfsck", "files",
                       "DOSFS check disk", rtems_shell_main_dosfsck);
  rtems_shell_add_cmd ("mkfsck", "files",
                       "EXT2FS format disk", rtems_ext2fs_mkfsck_main);
  rtems_shell_add_cmd ("bdbuftrace", "files",
                       "bdbuf trace toggle", shell_bdbuf_trace);
  rtems_shell_add_cmd ("td", "files",
                       "Test disk", disk_test_block_sizes);
#if RTEMS_RFS_TRACE
  rtems_shell_add_cmd ("rfs", "files",
                       "RFS trace",
                       rtems_rfs_trace_shell_command);
#endif
#if RTEMS_RFS_RTEMS_TRACE
  rtems_shell_add_cmd ("rrfs", "files",
                       "RTEMS RFS trace",
                       rtems_rfs_rtems_trace_shell_command);
#endif

#if RUN_TESTING
  rtems_rfs_bitmap_unit_test ();
#endif
  
  shell_init_script ();
  while (true)
    shell_start ();
  
  rtems_task_delete (RTEMS_SELF);

  return 0;
}

#if pc586
void
rtems_fatal_error_occurred (uint32_t code)
{
  printf ("fatal error: %08lx\n", code);
  for (;;);
}
#endif

/* Nonzero if either X or Y is not aligned on a "long" boundary.  */
#define UNALIGNED(X, Y) \
  (((long)X & (sizeof (long) - 1)) | ((long)Y & (sizeof (long) - 1)))

/* How many bytes are copied each iteration of the 4X unrolled loop.  */
#define BIGBLOCKSIZE    (sizeof (long) << 2)

/* How many bytes are copied each iteration of the word copy loop.  */
#define LITTLEBLOCKSIZE (sizeof (long))

/* Threshhold for punting to the byte copier.  */
#define TOO_SMALL(LEN)  ((LEN) < BIGBLOCKSIZE)
